"""Helper to facilitate the use of inputs and outputs of the models."""

# =============================================================================
# Copyright 2021 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import logging
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import torch

from ptlflow.data.flow_transforms import ToTensor
from ptlflow.models.base_model.base_model import BaseModel
from ptlflow.utils.utils import InputPadder, InputScaler


class IOAdapter(object):
    """Handle the inputs and outputs of optical flow models."""

    def __init__(
        self,
        model: BaseModel,
        input_size: Tuple[int, int],
        target_size: Optional[Tuple[int, int]] = None,
        target_scale_factor: Optional[float] = None,
        interpolation_mode: str = 'bilinear',
        interpolation_align_corners: bool = False,
        cuda: bool = False
    ) -> None:
        """Initialize IOAdapter.

        Parameters
        ----------
        model : BaseModel
            An instance of optical flow mode that will be used for estimation.
        input_size : Tuple[int, int]
            The shape of the original inputs, must be a tuple with at least two elements. It is assumed that the last two
            elements are (height, with).
        target_size : Optional[Tuple[int, int]], optional
            If provided, the inputs will be resized to target_size. target_size is defined as a tuple (height, with).
        target_scale_factor : Optional[float], default 1.0
            This value is only used if size is None. The multiplier that will be applied to the original shape to scale
            the input.
        interpolation_mode : str, default 'bilinear'
            How to perform the interpolation. It must be a value accepted by the 'mode' argument from
            torch.nn.functional.interpolate function.
        interpolation_align_corners : bool, default False
            Whether the interpolation keep the corners aligned. As defined in torch.nn.functional.interpolate.
        cuda : bool
            If True, the input tensors are transferred to GPU (if a GPU is available).
        """
        self.output_stride = model.output_stride
        self.target_size = target_size
        self.target_scale_factor = target_scale_factor
        self.interpolation_mode = interpolation_mode
        self.interpolation_align_corners = interpolation_align_corners
        self.cuda = cuda

        self.transform = ToTensor()
        padder_input_size = input_size
        self.scaler = None
        if (target_size is not None and min(target_size) > 0) or (target_scale_factor is not None and target_scale_factor > 0):
            self.scaler = InputScaler(
                orig_shape=input_size, size=target_size, scale_factor=target_scale_factor,
                interpolation_mode=interpolation_mode, interpolation_align_corners=interpolation_align_corners)

            if (target_size is not None and min(target_size) > 0):
                padder_input_size = target_size
            else:
                padder_input_size = (int(target_scale_factor*input_size[0]), int(target_scale_factor*input_size[1]))

        self.padder = InputPadder(padder_input_size, model.output_stride)

    def prepare_inputs(
        self,
        images: Optional[Union[np.ndarray, List[np.ndarray]]] = None,
        flows: Optional[Union[np.ndarray, List[np.ndarray]]] = None,
        inputs: Optional[Dict[str, Any]] = None,
        **kwargs: Union[np.ndarray, List[np.ndarray]]
    ) -> Dict[str, torch.Tensor]:
        """Transform numpy inputs into the input format of theoptical flow models.

        This basically consists on tranform the numpy arrays into torch tensors, and then putting them into a dict.

        Parameters
        ----------
        images : Union[np.ndarray, List[np.ndarray]]
            One or more images to use to estimate the optical flow. Typically, it will be a least with two images in the
            HWC format.
        flows : Optional[Union[np.ndarray, List[np.ndarray]]], optional
            One or more groundtruth optical flow, which can be used for validation. Typically it will be an array HWC.
        inputs : Optional[Dict[str, Any]]
            Dict containing input tensors or other metadata. Only the tensors will be transformed.
        kwargs : Union[np.ndarray, List[np.ndarray]]
            Any other array inputs can be provided as keyworded arguments. This function will create an entry in the input dict
            for each keyworded array given.

        Returns
        -------
        Dict[str, Any]
            The inputs converted and transformed to the input format of the optical flow models.
        """
        if inputs is None:
            inputs = {'images': images, 'flows': flows}
            inputs.update(kwargs)
            keys_to_remove = []
            for k, v in inputs.items():
                if v is None or len(v) == 0:
                    keys_to_remove.append(k)
            for k in keys_to_remove:
                del inputs[k]
            inputs = self.transform(inputs)

        for k, v in inputs.items():
            if isinstance(v, torch.Tensor):
                while len(v.shape) < 5:
                    v = v.unsqueeze(0)
                if self.scaler is not None:
                    v = self.scaler.scale(v, is_flow=k.startswith('flow'))
                v = self.padder.pad(v)
                inputs[k] = v

        inputs = self._to_cuda(inputs)

        return inputs

    def unpad_and_unscale(
        self,
        outputs: Dict[str, Any]
    ) -> Dict[str, torch.Tensor]:
        """Remove padding and scaling to restore the original shapes.

        The inputs and outputs obtained from prepare_inputs() may have been padded and scaled.
        This function can be used to revert those operations.

        Typically, this function should be used on both the inputs and outputs of the model after the model generated the
        predictions.

        Parameters
        ----------
        outputs : Dict[str, Any]
            It can be either the inputs or the outputs of optical flow models. All tensors will have the extra padding and
            rescaling removed.

        Returns
        -------
        Dict[str, Any]
            The same tensors, with the padding and scaling removed.

        See Also
        --------
        prepare_inputs
        """
        for k, v in outputs.items():
            if isinstance(v, torch.Tensor):
                v = self.padder.unpad(v)
                if self.scaler is not None:
                    v = self.scaler.unscale(v, is_flow=k.startswith('flow'))
                outputs[k] = v

        return outputs

    def _to_cuda(
        self,
        inputs: Dict[str, Any]
    ) -> Dict[str, Any]:
        if self.cuda:
            if torch.cuda.is_available():
                inputs = {k: v.cuda() if isinstance(v, torch.Tensor) else v for k, v in inputs.items()}
            else:
                logging.warning(
                    'IOAdapter was asked to use cuda, but torch.cuda.is_available() == False. Tensors will remain on CPU.')
        return inputs



from argparse import ArgumentParser

try:
    from spatial_correlation_sampler import SpatialCorrelationSampler
except ModuleNotFoundError:
    from ptlflow.utils.correlation import IterSpatialCorrelationSampler as SpatialCorrelationSampler
import torch
import torch.nn as nn

from .pwc_modules import conv, upsample2d_as, rescale_flow, initialize_msra
from .pwc_modules import WarpingLayer, FeatureExtractor
from .pwc_modules import FlowAndOccContextNetwork, FlowAndOccEstimatorDense
from .irr_modules import OccUpsampleNetwork, RefineFlow, RefineOcc
from ptlflow.models.base_model.base_model import BaseModel


class StarFlow(BaseModel):
    pretrained_checkpoints = {
        'things': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/starflow-things-d3966a93.ckpt',
        'sintel': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/starflow-sintel-3c06b844.ckpt',
        'kitti': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/starflow-kitti-34b9a3ed.ckpt'
    }

    def __init__(self, args):
        super(StarFlow, self).__init__(
            args=args,
            loss_fn=None,
            output_stride=64)
        self.args = args
        self.leakyRELU = nn.LeakyReLU(0.1, inplace=True)

        self.feature_pyramid_extractor = FeatureExtractor(self.args.num_chs)
        self.warping_layer = WarpingLayer()

        self.dim_corr = (self.args.search_range * 2 + 1) ** 2
        self.num_ch_in = self.dim_corr + 32 + 2 + 1

        self.flow_and_occ_estimators = FlowAndOccEstimatorDense(2 * self.num_ch_in)
        self.context_networks = FlowAndOccContextNetwork(2 * self.num_ch_in + 448 + 2 + 1)

        self.occ_shuffle_upsample = OccUpsampleNetwork(11, 1)

        self.conv_1x1 = nn.ModuleList([conv(196, 32, kernel_size=1, stride=1, dilation=1),
                                       conv(128, 32, kernel_size=1, stride=1, dilation=1),
                                       conv(96, 32, kernel_size=1, stride=1, dilation=1),
                                       conv(64, 32, kernel_size=1, stride=1, dilation=1)])

        self.conv_1x1_1 = conv(16, 3, kernel_size=1, stride=1, dilation=1)

        self.conv_1x1_time = conv(2 * self.num_ch_in + 448, self.num_ch_in, kernel_size=1, stride=1, dilation=1)

        self.refine_flow = RefineFlow(2 + 1 + 32)
        self.refine_occ = RefineOcc(1 + 32 + 32)

        self.corr = SpatialCorrelationSampler(kernel_size=1, patch_size=2*self.args.search_range+1, padding=0)

        initialize_msra(self.modules())

    @staticmethod
    def add_model_specific_args(parent_parser=None):
        parent_parser = BaseModel.add_model_specific_args(parent_parser)
        parser = ArgumentParser(parents=[parent_parser], add_help=False)
        parser.add_argument('--div_flow', type=float, default=0.05)
        parser.add_argument('--search_range', type=int, default=4)
        parser.add_argument('--output_level', type=int, default=4)
        parser.add_argument('--num_chs', type=int, nargs='+', default=[3, 16, 32, 64, 96, 128, 196])
        parser.add_argument('--num_iters', type=int, default=1)
        parser.add_argument('--num_levels', type=int, default=7)
        return parser

    def forward(self, inputs):
        images = inputs['images']
        list_imgs = [images[:, i] for i in range(images.shape[1])]

        _, _, height_im, width_im = list_imgs[0].size()

        # on the bottom level are original images
        list_pyramids = [] #indices : [time][level]
        for im in list_imgs:
            list_pyramids.append(self.feature_pyramid_extractor(im) + [im])

        # outputs
        output_dict = {}
        output_dict_eval = {}
        flows_f = [] #indices : [level][time]
        flows_b = [] #indices : [level][time]
        occs_f = []
        occs_b = []
        flows_coarse_f = []
        occs_coarse_f = []
        for l in range(len(list_pyramids[0])):
            flows_f.append([])
            flows_b.append([])
            occs_f.append([])
            occs_b.append([])
        for l in range(self.args.output_level + 1):
            flows_coarse_f.append([])
            occs_coarse_f.append([])

        # init
        b_size, _, h_x1, w_x1, = list_pyramids[0][0].size()
        init_dtype = list_pyramids[0][0].dtype
        init_device = list_pyramids[0][0].device
        flow_f = torch.zeros(b_size, 2, h_x1, w_x1, dtype=init_dtype, device=init_device).float()
        flow_b = torch.zeros(b_size, 2, h_x1, w_x1, dtype=init_dtype, device=init_device).float()
        occ_f = torch.zeros(b_size, 1, h_x1, w_x1, dtype=init_dtype, device=init_device).float()
        occ_b = torch.zeros(b_size, 1, h_x1, w_x1, dtype=init_dtype, device=init_device).float()
        previous_features = []

        for i in range(len(list_imgs) - 1):
            x1_pyramid, x2_pyramid = list_pyramids[i:i+2]

            for l, (x1, x2) in enumerate(zip(x1_pyramid, x2_pyramid)):

                if l <= self.args.output_level:
                    if i == 0:
                        bs_, _, h_, w_, = list_pyramids[0][l].size()
                        previous_features.append(torch.zeros(bs_, self.num_ch_in, h_, w_, dtype=init_dtype, device=init_device).float())

                    # warping
                    if l == 0:
                        x2_warp = x2
                        x1_warp = x1
                    else:
                        flow_f = upsample2d_as(flow_f, x1, mode="bilinear")
                        flow_b = upsample2d_as(flow_b, x2, mode="bilinear")
                        occ_f = upsample2d_as(occ_f, x1, mode="bilinear")
                        occ_b = upsample2d_as(occ_b, x2, mode="bilinear")
                        x2_warp = self.warping_layer(x2, flow_f, height_im, width_im, self.args.div_flow)
                        x1_warp = self.warping_layer(x1, flow_b, height_im, width_im, self.args.div_flow)

                    # correlation
                    out_corr_f = self.corr(x1, x2_warp)
                    out_corr_f = out_corr_f.view(out_corr_f.shape[0], -1, out_corr_f.shape[3], out_corr_f.shape[4])
                    out_corr_f = out_corr_f / x1.shape[1]

                    out_corr_b = self.corr(x2, x1_warp)
                    out_corr_b = out_corr_b.view(out_corr_b.shape[0], -1, out_corr_b.shape[3], out_corr_b.shape[4])
                    out_corr_b = out_corr_b / x2.shape[1]

                    out_corr_relu_f = self.leakyRELU(out_corr_f)
                    out_corr_relu_b = self.leakyRELU(out_corr_b)

                    if l != self.args.output_level:
                        x1_1by1 = self.conv_1x1[l](x1)
                        x2_1by1 = self.conv_1x1[l](x2)
                    else:
                        x1_1by1 = x1
                        x2_1by1 = x2

                    if i > 0: #temporal connection:
                        previous_features[l] = self.warping_layer(previous_features[l],
                                        flows_b[l][-1], height_im, width_im, self.args.div_flow)

                    # Flow and occlusions estimation
                    flow_f = rescale_flow(flow_f, self.args.div_flow, width_im, height_im, to_local=True)
                    flow_b = rescale_flow(flow_b, self.args.div_flow, width_im, height_im, to_local=True)

                    features = torch.cat([previous_features[l], out_corr_relu_f, x1_1by1, flow_f, occ_f], 1)
                    features_b = torch.cat([torch.zeros_like(previous_features[l]), out_corr_relu_b, x2_1by1, flow_b, occ_b], 1)

                    x_intm_f, flow_res_f, occ_res_f = self.flow_and_occ_estimators(features)
                    flow_est_f = flow_f + flow_res_f
                    occ_est_f = occ_f + occ_res_f
                    with torch.no_grad():
                        x_intm_b, flow_res_b, occ_res_b = self.flow_and_occ_estimators(features_b)
                        flow_est_b = flow_b + flow_res_b
                        occ_est_b = occ_b + occ_res_b

                    # Context:
                    flow_cont_res_f, occ_cont_res_f = self.context_networks(torch.cat([x_intm_f, flow_est_f, occ_est_f], dim=1))
                    flow_cont_f = flow_est_f + flow_cont_res_f
                    occ_cont_f = occ_est_f + occ_cont_res_f
                    with torch.no_grad():
                        flow_cont_res_b, occ_cont_res_b = self.context_networks(torch.cat([x_intm_b, flow_est_b, occ_est_b], dim=1))
                        flow_cont_b = flow_est_b + flow_cont_res_b
                        occ_cont_b = occ_est_b + occ_cont_res_b

                    # refinement
                    img1_resize = upsample2d_as(list_imgs[i], flow_f, mode="bilinear")
                    img2_resize = upsample2d_as(list_imgs[i+1], flow_b, mode="bilinear")
                    flow_cont_f = rescale_flow(flow_cont_f, self.args.div_flow, width_im, height_im, to_local=False)
                    flow_cont_b = rescale_flow(flow_cont_b, self.args.div_flow, width_im, height_im, to_local=False)
                    img2_warp = self.warping_layer(img2_resize, flow_cont_f, height_im, width_im, self.args.div_flow)
                    img1_warp = self.warping_layer(img1_resize, flow_cont_b, height_im, width_im, self.args.div_flow)

                    # flow refine
                    flow_f = self.refine_flow(flow_cont_f.detach(), img1_resize - img2_warp, x1_1by1)
                    flow_b = self.refine_flow(flow_cont_b.detach(), img2_resize - img1_warp, x2_1by1)

                    flow_f = rescale_flow(flow_f, self.args.div_flow, width_im, height_im, to_local=False)
                    flow_b = rescale_flow(flow_b, self.args.div_flow, width_im, height_im, to_local=False)

                    # occ refine
                    x2_1by1_warp = self.warping_layer(x2_1by1, flow_f, height_im, width_im, self.args.div_flow)
                    x1_1by1_warp = self.warping_layer(x1_1by1, flow_b, height_im, width_im, self.args.div_flow)

                    occ_f = self.refine_occ(occ_cont_f.detach(), x1_1by1, x1_1by1 - x2_1by1_warp)
                    occ_b = self.refine_occ(occ_cont_b.detach(), x2_1by1, x2_1by1 - x1_1by1_warp)

                    # save features for temporal connection:
                    previous_features[l] = self.conv_1x1_time(x_intm_f)
                    flows_f[l].append(flow_f)
                    occs_f[l].append(occ_f)
                    flows_b[l].append(flow_b)
                    occs_b[l].append(occ_b)
                    flows_coarse_f[l].append(flow_cont_f)
                    occs_coarse_f[l].append(occ_cont_f)
                    #flows.append([flow_cont_f, flow_cont_b, flow_f, flow_b])
                    #occs.append([occ_cont_f, occ_cont_b, occ_f, occ_b])

                else:
                    flow_f = upsample2d_as(flow_f, x1, mode="bilinear")
                    flow_b = upsample2d_as(flow_b, x2, mode="bilinear")
                    flows_f[l].append(flow_f)
                    flows_b[l].append(flow_b)
                    #flows.append([flow_f, flow_b])

                    x2_warp = self.warping_layer(x2, flow_f, height_im, width_im, self.args.div_flow)
                    x1_warp = self.warping_layer(x1, flow_b, height_im, width_im, self.args.div_flow)
                    flow_b_warp = self.warping_layer(flow_b, flow_f, height_im, width_im, self.args.div_flow)
                    flow_f_warp = self.warping_layer(flow_f, flow_b, height_im, width_im, self.args.div_flow)

                    if l != self.args.num_levels-1:
                        x1_in = self.conv_1x1_1(x1)
                        x2_in = self.conv_1x1_1(x2)
                        x1_w_in = self.conv_1x1_1(x1_warp)
                        x2_w_in = self.conv_1x1_1(x2_warp)
                    else:
                        x1_in = x1
                        x2_in = x2
                        x1_w_in = x1_warp
                        x2_w_in = x2_warp

                    occ_f = self.occ_shuffle_upsample(occ_f, torch.cat([x1_in, x2_w_in, flow_f, flow_b_warp], dim=1))
                    occ_b = self.occ_shuffle_upsample(occ_b, torch.cat([x2_in, x1_w_in, flow_b, flow_f_warp], dim=1))

                    occs_f[l].append(occ_f)
                    occs_b[l].append(occ_b)
                    #occs.append([occ_f, occ_b])

            flow_f = torch.zeros(b_size, 2, h_x1, w_x1, dtype=init_dtype, device=init_device).float()
            flow_b = torch.zeros(b_size, 2, h_x1, w_x1, dtype=init_dtype, device=init_device).float()
            occ_f = torch.zeros(b_size, 1, h_x1, w_x1, dtype=init_dtype, device=init_device).float()
            occ_b = torch.zeros(b_size, 1, h_x1, w_x1, dtype=init_dtype, device=init_device).float()

        outputs = {}


        if self.training:
            raise NotImplementedError('Training is still not implemented for StarFlow.')
        else:
            outputs['flows'] = torch.stack(
                [upsample2d_as(f, list_imgs[0], mode="bilinear") / self.args.div_flow for f in flows_f[-1]], dim=1)
            outputs['occs'] = torch.stack(
                [upsample2d_as(torch.sigmoid(o), list_imgs[0], mode="bilinear") for o in occs_f[-1]], dim=1)
            outputs['flows_b'] = torch.stack(
                [upsample2d_as(f, list_imgs[0], mode="bilinear") / self.args.div_flow for f in flows_b[-1]], dim=1)
            outputs['occs_b'] = torch.stack(
                [upsample2d_as(torch.sigmoid(o), list_imgs[0], mode="bilinear") for o in occs_b[-1]], dim=1)

        return outputs